import numpy as np
import xarray as xr

import sectionate as sec
from sectionate import is_section_counterclockwise
from .utilities import *
from .grid_conform import (
    get_region_boundary_grid_indices,
    mask_from_grid_boundaries
)

import os
from pathlib import Path

class Region:
    """
    A named polygonal region defined by a list or array of geographical coordinates.
    """
    def __init__(
        self,
        name,
        lons,
        lats,
        force_ccw=False,
        remove_duplicate_points=False
        ):
        """
        Create a Region object (named `name`) from arrays of (`lons`, `lats`).

        PARAMETERS
        ----------
        name : str
            Name of the region
        lons : list or np.ndarray
            Longitudes (in degrees).
        lats : list or np.ndarray
            Latitudes (in degrees).
        force_ccw : bool
            Default: False. If True, checks if Region is clockwise and, if it is,
            swaps the order of the points that define it such that it becomes counterclockwise.
        remove_duplicate_points : bool
            Default: False. If True, prunes any duplicate points from the input arrays (lons, lats).

        RETURNS
        -------
        Region instance

        Examples
        --------
        >>> lons, lats = np.array([-80., -66., -65.]), np.array([ 26.,  18.,  32.])
        >>> region = reg.Region('Bermuda Triangle', lons, lats)
        """
        
        self.name = name
        self.lons_c = lons
        self.lats_c = lats
        
        if remove_duplicate_points:
            self.remove_duplicate_points()

        self.counterclockwise = is_section_counterclockwise(
            loop(self.lons_c),
            loop(self.lats_c),
            geometry='spherical'
        )
            
        if force_ccw:
            self.make_counterclockwise()

    def copy(self, remove_duplicate_points=False):
        """
        Returns a copy of the Region.

        PARAMETERS
        ----------
        remove_duplicate_points : bool
            Default: False. If True, prunes any duplicate points from the input arrays (lons, lats).
            
        RETURNS
        ----------
        region_copy : regionate.region.Region
            Copy of the region.
        """
        return Region(
            self.name,
            self.lons_c.copy(),
            self.lats_c.copy(),
            remove_duplicate_points=remove_duplicate_points
        )
    
    def make_counterclockwise(self):
        """
        Checks if the section is clockwise and flips its direction if it is to make it counterclockwise.
        """
        if not(self.counterclockwise):
            self.lons_c = self.lons_c[::-1]
            self.lats_c = self.lats_c[::-1]
            self.counterclockwise = True

    def remove_duplicate_points(self, closeness_threshold=5.e3):
        """
        Removes any duplicate points.
        
        PARAMETERS
        ----------
        closeness_threshold : float
            A short distance within which points are deemed to be identical. Default: 5.e3.
        """
        self.lons_c, self.lats_c = unique_lonlat(
            self.lons_c,
            self.lats_c,
            closeness_threshold=closeness_threshold
        )

    def __repr__(self):
        return f"{str(type(self))[8:-2]}('{self.name}')"
        
class GriddedRegion(Region):
    """
    A named polygonal region that exactly conforms to the velocity faces of a C-grid ocean model.
    """
    def __init__(
        self,
        name,
        lons,
        lats,
        grid,
        positive_in=True,
        mask=None,
        ij=None,
        ):
        """
        Create a Region object (named `name`) from arrays of (`lons`, `lats`) and an ocean model `grid`. 

        PARAMETERS
        ----------
        name : str
            Name of the region
        lons : list or np.ndarray
            Longitudes (in degrees).
        lats : list or np.ndarray
            Latitudes (in degrees).
        grid : xgcm.Grid
        positive_in : bool
            Default: True. If True, prunes any duplicate points from the input arrays (lons, lats).
        mask : None or xr.DataArray (default: None)
            If None, does not apply any mask.
        ij : None or list
            If None, the indices of grid coordinates closest to provided coordinates `self.i_c` and `self.j_c`
            are inferred from the model grid. If list, assume two elements in the list and
            extract `self.i_c = ij[0]` and `self.j_c = ij[1]`.

        RETURNS
        -------
        GriddedRegion instance

        Examples
        --------
        >>> grid = xgcm.Grid(...) # TO DO: minimal example grid
        >>> lons, lats = np.array([-80., -66., -65.]), np.array([ 26.,  18.,  32.])
        >>> region = reg.Region('Bermuda Triangle', lons, lats, grid)
        """

        if any([c not in grid._ds.coords for c in ["geolon_c", "geolat_c"]]):
            raise ValueError("grid._ds must contain coordinates of grid cell corners, named 'geolon_c' and 'geolat_c'.")

        self.grid = grid
        self.save = {}
        
        if len(lons)>=3 and len(lats)>=3 and ij is None:
            self.initiate_from_boundary(
                lons,
                lats,
                mask=mask,
                positive_in=positive_in
            )
        elif ij is None:
            raise NameError("Must provide lons and lats as lists or arrays\
            to define the region.")
        else:
            self.lons_c = lons
            self.lats_c = lats
            self.i_c = ij[0]
            self.j_c = ij[1]
            if mask is None:
                self.mask = mask_from_grid_boundaries(
                    self.lons_c,
                    self.lats_c,
                    self.grid,
                    along_boundary=True
                )
            else:
                self.mask = mask
        
        super().__init__(
            name=name,
            lons=self.lons_c,
            lats=self.lats_c
        )
        
    def initiate_from_boundary(
        self,
        lons,
        lats,
        positive_in=True,
        mask=None
        ):
        """
        TO DO
        """

        self.i_c, self.j_c, self.lons_c, self.lats_c, self.lons_uv, self.lats_uv = (
            get_region_boundary_grid_indices(
                lons.copy(),
                lats.copy(),
                self.grid
            )
        )
        if mask is None:
            mask = mask_from_grid_boundaries(
                self.lons_c,
                self.lats_c,
                self.grid
            )
        self.mask = mask.astype(bool) ^ (not positive_in)

    def to_gr(self, path):
        """Save the GriddedRegion object in a .gr format file directory

        There are two key files within each .gr directory:
          - a `grid.nc` file that contains information about the coordinates
          requires to create an `xgcm.Grid` instance
          - a `boundary.nc` file that contains the region's tracer cell mask
          and the coordinates and indices of the corner cells that define its
          boundary.

        To do:
          - Subdirectory for child boundary information

        Arguments
        ---------
        path [str] -- path to directory where the .gr file directory
            should be saved. The filename will be [GriddedRegion.name].gr

        Example
        -------
        >>> gridded_region.save('../data/')
        """
        gr_path = f"{path}/{self.name.replace(' ','_')}.gr/"
        Path(gr_path).mkdir(parents=True, exist_ok=True)

        grid_path = f"{gr_path}/grid.nc"
        parent_grid = f"{path}/../grid.nc"
        if os.path.isfile(parent_grid):
            os.symlink(parent_grid, grid_path)
        else:
            grid = self.grid
            grid._ds.drop_vars([v for v in grid._ds.data_vars]).to_netcdf(grid_path)

        # Write boundary information to NetCDF file
        vertex = xr.DataArray(np.arange(0, self.i_c.size), dims=('vertex',))
        face = xr.DataArray(np.arange(0.5, self.i_c.size-1), dims=('face',))
        ds = xr.Dataset({}, coords={'vertex': vertex, 'face': face})
        for v in ['lons', 'lats', 'i', 'j']:
            var = getattr(self,v)
            if v in ['lons', 'lats']:
                var = loop(var)
            ds[v] = xr.DataArray(var, dims=('vertex',))
        for v in ['lons_uv', 'lats_uv']:
            ds[v] = xr.DataArray(getattr(self,v), dims=('face',))
        ds['mask'] = self.mask
        ds.to_netcdf(f"{gr_path}/region.nc")

        for (k,v) in self.save.items():
            v.to_netcdf(f"{gr_path}/{k}.nc")

        # Write boundary information for each child section to NetCDF file
        child_path = f"{gr_path}/children/"
        Path(child_path).mkdir(parents=True, exist_ok=True)
        for child in self.children.values():
            child_name = child.name.replace(' ','_')
            sec_path = f"{gr_path}/children/{child_name}.sec"
            Path(sec_path).mkdir(parents=True, exist_ok=True)
            
            vertex = xr.DataArray(np.arange(0, child.i_c.size), dims=('vertex',))
            ds = xr.Dataset({}, coords={'vertex': vertex})
            for k in ['lons', 'lats', 'i', 'j']:
                var = getattr(child,k)
                ds[k] = xr.DataArray(var, dims=('vertex',))
            ds.to_netcdf(f"{sec_path}/section.nc")
            for k, ds_save in child.save.items():
                ds_save.to_netcdf(f"{sec_path}/{k}.nc")

class BoundedRegion(GriddedRegion):
    def __init__(self, section, grid, **kwargs):
        super().__init__(
            section.name,
            section.lons_c,
            section.lats_c,
            grid,
            **kwargs
        )
        self.children = {}

        def slice_indices_in_list(s, l):
            """
            Return a list of indices in `l` where elements of slice `s` appear in order.
            Returns list of indices if found, else None.
            """
            len_s = len(s)
            for i in range(len(l) - len_s + 1):
                if s == l[i:i+len_s]:
                    return np.array(list(range(i, i+len_s)), dtype=int)
            return None

        # Create the parent gridded section
        parent_section_gridded = sec.GriddedSection(
            sec.Section(
                section.name,
                (self.lons_c, self.lats_c),
            ),
            grid,
            i_c = self.i_c,
            j_c = self.j_c
        )
        parent_lons_uv, parent_lats_uv = sec.uvcoords_from_qindices(
            grid,
            parent_section_gridded.i_c,
            parent_section_gridded.j_c
        )
        parent_coords_uv = sec.coords_from_lonlat(parent_lons_uv, parent_lats_uv)
            
        for child_name, child in section.children.items():
            i_c, j_c, lons_c, lats_c = sec.grid_section(grid, child.lons_c, child.lats_c)
            
            child_coords = sec.coords_from_lonlat(lons_c, lats_c)

            # Find the indices in the parent's corner sections that correspond this child
            parent_idx_c = slice_indices_in_list(child_coords, parent_section_gridded.coords)
            if parent_idx_c is not None:
                pass  # child orientation matches parent
            else:
                parent_idx_c = slice_indices_in_list(child_coords[::-1], parent_section_gridded.coords)
                if parent_idx_c is not None:
                    # reversed slice is in the parent list
                    child.lons_c = child.lons_c[::-1]
                    child.lats_c = child.lats_c[::-1]
                    # recompute the child sections using the correct orientation
                    i_c, j_c, lons_c, lats_c = sec.grid_section(grid, child.lons_c, child.lats_c)
                else:
                    raise ValueError("Child corner sections do not match up with parent ones!")
            
            # Find the indices in the parent's velocity sections that correspond this child
            lons_uv, lats_uv = sec.uvcoords_from_qindices(grid, i_c, j_c)
            child_coords_uv = sec.coords_from_lonlat(lons_uv, lats_uv)
            parent_idx_uv = slice_indices_in_list(child_coords_uv, parent_coords_uv)
            if parent_idx_uv is None:
                raise ValueError("Child velocity sections do not match up with parent ones!")

            child_coords = sec.coords_from_lonlat(lons_c, lats_c)
            child_section = sec.Section(
                child_name,
                child_coords,
                children={},
                parent=parent_section_gridded
            )
            child_section_gridded = sec.GriddedSection(
                child_section,
                grid,
                i_c = i_c,
                j_c = j_c,
            )
            child_section_gridded.parent_idx_c = parent_idx_c
            child_section_gridded.parent_idx_uv = parent_idx_uv
            self.children[child_name] = child_section_gridded

def open_gr(path, ds_to_grid):

    ds_grid = xr.open_dataset(f"{path}/grid.nc")
    grid = ds_to_grid(ds_grid)
    ds = xr.open_dataset(f"{path}/region.nc")
    
    name = path.split('/')[-1][:-3].replace('_',' ')
    region = GriddedRegion(
        name,
        ds.lons_c.values,
        ds.lats_c.values,
        grid,
        mask = ds.mask,
        ij = (ds.i_c.values, ds.j_c.values)
    )
    gr_files = [
        f for f in os.listdir(f"{path}/")
        if ('.nc' in f) and (f not in ['grid.nc', 'region.nc'])
    ]
    for file in gr_files:
        v = file.split('.')[0]
        region.save[v] = xr.open_dataset(f"{path}/{file}")

    region.children = {}
    children_path = f"{path}/children/"
    child_paths = [
        f"{children_path}{file}"
        for file in os.listdir(children_path)
    ]
    for child_path in child_paths:
        child_name = child_path.split('/')[-1][:-4].replace('_', ' ')
        ds = xr.open_dataset(f"{child_path}/section.nc")
        
        section = sec.Section(
            child_name,
            sec.coords_from_lonlat(ds.lons_c.values, ds.lats_c.values)
        )
        
        section.save = {}
        for file in [f for f in os.listdir(child_path) if f != 'section.nc']:
            v = file.split('.')[0]
            section.save[v] = xr.open_dataset(f"{child_path}/{file}")

        region.children[child_name] = section
        
    return region


import numpy as np
import xarray as xr
import xgcm

def initialize_spherical_grid(N=6):
    xq = np.linspace(0, 360., N+1)
    yq = np.linspace(-60, 60., N+1)
    dx = 360/N
    xh = np.linspace(0+dx/2, 360-dx/2, N)
    dy = 120/N
    yh = np.linspace(-60+dy/2, 60-dy/2, N)

    lon, lat = np.meshgrid(xh, yh)
    lon_c, lat_c = np.meshgrid(xq, yq)
    ds = xr.Dataset({}, coords={
        "xh":xr.DataArray(xh, dims=("xh",)),
        "yh":xr.DataArray(yh, dims=("yh",)),
        "xq":xr.DataArray(xq, dims=("xq",)),
        "yq":xr.DataArray(yq, dims=("yq",)),
        "geolon":xr.DataArray(lon, dims=("yh", "xh")),
        "geolat":xr.DataArray(lat, dims=("yh", "xh")),
        "geolon_c":xr.DataArray(lon_c, dims=("yq", "xq",)),
        "geolat_c":xr.DataArray(lat_c, dims=("yq", "xq",))
    })
    coords = {
        'X': {'outer': 'xq', 'center': 'xh'},
        'Y': {'outer': 'yq', 'center': 'yh'}
    }
    grid = xgcm.Grid(ds, coords=coords, boundary={"X":"periodic", "Y":"extend"}, autoparse_metadata=False)
    return grid

def test_gridded_region_from_boundary():
    from regionate import GriddedRegion
    from sectionate import distance_on_unit_sphere

    lonseg = np.array([0., 120., 240, 360.])
    latseg = np.array([0.,   0.,   0.,  0.])

    grid = initialize_spherical_grid()
    region = GriddedRegion("test_region1", lonseg, latseg, grid)

    dists = distance_on_unit_sphere(
        region.lons_c,
        region.lats_c,
        np.array([0.,  60., 120., 180., 240., 300., 360.]),
        np.array([0.,   0.,   0.,   0.,   0.,   0.,   0.])
    )
    assert np.all(np.isclose(dists, 0., atol=1.e-6))

    region_rev = GriddedRegion("test_region2", lonseg[::-1], latseg[::-1], grid)
    assert np.all(np.equal(region.mask, region_rev.mask))
    
def test_gridded_region_from_mask():
    from regionate import MaskRegions
    
    grid = initialize_spherical_grid()
    
    # Two grid-cell wide interior region mask
    mask = xr.ones_like(grid._ds.geolon).where((grid._ds.xh==90.) & (np.abs(grid._ds.yh)<=10), 0.).astype(bool)
    region_dict = MaskRegions(mask, grid).region_dict
    assert len(region_dict)==1
    
    region = region_dict[0]
    assert np.all(
        modequal(region.lons_c, np.array([ 60., 120., 120., 120.,  60.,  60.])) &
        modequal(region.lats_c, np.array([-20., -20.,   0.,  20.,  20.,   0.]))
    )
    
    # Zonal strip mask
    mask = xr.ones_like(grid._ds.geolon).where(np.abs(grid._ds.yh)<=10, 0.).astype(bool)
    region = MaskRegions(mask, grid).region_dict[0]
    assert np.all(
        modequal(region.lons_c, np.array([360.,  60., 120., 180., 240., 300., 360., 360., 360., 300., 240., 180., 120.,  60., 360., 360.])) &
        modequal(region.lats_c, np.array([-20., -20., -20., -20., -20., -20., -20.,   0.,  20.,  20.,  20., 20.,  20.,  20.,  20.,   0.]))
    )
    
    # All but zonal strip mask (two separate regions outside)
    region_inv = MaskRegions(~mask, grid).region_dict
    assert np.all(
        modequal(region_inv[0].lons_c, np.array([360.,  60., 120., 180., 240., 300., 360., 360., 360., 300., 240., 180., 120.,  60., 360., 360.])) &
        modequal(region_inv[0].lats_c, np.array([-60., -60., -60., -60., -60., -60., -60., -40., -20., -20., -20., -20., -20., -20., -20., -40.])) &
        modequal(region_inv[1].lons_c, np.array([360.,  60., 120., 180., 240., 300., 360., 360., 360., 300., 240., 180., 120.,  60., 360., 360.])) &
        modequal(region_inv[1].lats_c, np.array([20.,   20.,  20.,  20.,  20.,  20.,  20.,  40.,  60.,  60.,  60.,  60.,  60.,  60.,  60.,  40.]))
    )

def modequal(a,b):
    return np.equal(np.mod(a, 360.), np.mod(b, 360.))


import numpy as np
from sectionate import distance_on_unit_sphere

def split_non_consecutive_list(data, mod=np.inf):
    """TO DO
    
    See also
    --------
    consecutive_lists, wrap_non_consecutive_listsoflists
    """
    if len(data) == 0:
        return [[]]
    else:
        data = iter(np.array(data).astype("int64"))
        val = next(data)
        chunk = []
        try:
            while True:
                chunk.append(val)
                val = next(data)
                if val != np.mod(chunk[-1] + 1, mod):
                    yield chunk
                    chunk = []
        except StopIteration:
            if chunk:
                yield chunk

def wrap_non_consecutive_listsoflists(lol, mod=np.inf):
    """TO DO"""
    if lol[0][0] == np.mod(lol[-1][-1]+1, mod):
        if len(lol)==1:
            lol[0] = lol[0][1:]
        else:
            lol = [lol[-1]+lol[0]] + lol[1:-1]
    lol = [l for l in lol if len(l)>1]
    return lol

def consecutive_lists(data, mod=np.inf):
    """TO DO"""
    if any(data):
        lol = [i for i in split_non_consecutive_list(data, mod=mod)]
        if any(lol):
            return wrap_non_consecutive_listsoflists(lol, mod=mod)
        else:
            return [[]]
    else:
        return [[]]
    
def coord_list(lons, lats):
    """Turns iterable longitudes and latitudes into a list of coordinate pairs"""
    return [(lon, lat) for (lon, lat) in zip(lons, lats)]
    
def unique_list(l):
    """Removes duplicate entries from list"""
    u = []
    for i, e in enumerate(l):
        # check if exists in unique_list or not
        if e not in u:
            u.append(e)
    return u

def unique_coords(coords, closeness_threshold=5.e3):
    """Removes duplicate coordinates from a list of (lon, lat) coordinate pairs

    PARAMETERS
    ----------
    coords : list of (lon, lat) coordinate pairs
    closeness_threshold : cutoff distance between "identical" points, as float (default: 5.e3)

    RETURNS
    -------
    coords

    See also
    --------
    unique_lonlat
    """
    lons = [coord[0] for coord in coords]
    lats = [coord[1] for coord in coords]
    uc = []
    for i, coord in enumerate(coords):
        # check if exists in unique_list or not
        if not(np.any( distance_on_unit_sphere(coord[0], coord[1], [c[0] for c in uc], [c[1] for c in uc]) < closeness_threshold )):
            uc.append(coord)
    return uc

def unique_lonlat(lons, lats, closeness_threshold=5.e3):
    """Removes duplicate coordinates from geographical coordinate arrays

    PARAMETERS
    ----------
    lons : array of longitudes in degrees
    lats : array of latitudes in degrees
    closeness_threshold : cutoff distance between "identical" points, as float (default: 5.e3)

    RETURNS
    -------
    lons, lats

    See also
    --------
    unique_coords
    """
    uc = unique_coords(coord_list(lons, lats), closeness_threshold=closeness_threshold)
    return np.array([c[0] for c in uc]), np.array([c[1] for c in uc])

def lon_mod(lon, lon_ref):
    """Longitudes modulo 360 degrees

    PARAMETERS
    ----------
    lon : input numpy array of longitudes [in degrees]
    lon_ref : reference longitude [in degrees]
    
    RETURNS
    -------
    np.ndarray with shape of `lon` but modulo 360 degrees
    
    """
    return lon + 360. *np.round((lon_ref - lon)/360.)

def loop(x):
    """Loop a list or numpy array by appending the first value to the end."""
    return np.append(x, x[0])

def flatten_dict_values(d):
    """
    Flattens iterable dictionary values into a single list.
    
    See also
    --------
    flatten_lol
    """
    return flatten_lol(d.values)

def flatten_lol(lol):
    """Flattens a list of lists into a single list"""
    return [i for l in lol for i in l]


import numpy as np
from .utilities import *

def find_indices_of_overlaps(r1, r2, closeness_threshold=5.e3, face_indices=False):
    overlaps = distance_on_unit_sphere(
            r1.lons_c[:, np.newaxis], r1.lats_c[:, np.newaxis],
            r2.lons_c[np.newaxis, :], r2.lats_c[np.newaxis, :]
        ) < closeness_threshold

    r1_oidx = consecutive_lists( np.where(np.any(overlaps, axis=1))[0], mod=r1.lons_c.size)
    r2_oidx = consecutive_lists( np.where(np.any(overlaps, axis=0))[0], mod=r2.lons_c.size)
    
    if face_indices:
        r1_oidx = [l[:-1] for l in r1_oidx]
        r2_oidx = [l[:-1] for l in r2_oidx]
    
    return group_overlaps({r1.name: r1_oidx, r2.name: r2_oidx}, r1, r2, closeness_threshold=closeness_threshold)

def group_overlaps(overlaps, r1, r2, closeness_threshold=5.e3):
    grouped_overlaps = {r1.name: {}, r2.name: {}}
    group_num = 0
    for o2 in overlaps[r2.name]:
        for o1 in overlaps[r1.name]:
            if np.any(o1) and np.any(o2):
                if np.any(distance_on_unit_sphere(
                        r1.lons_c[o1, np.newaxis], r1.lats_c[o1, np.newaxis],
                        r2.lons_c[np.newaxis, o2], r2.lats_c[np.newaxis, o2]
                    ) < closeness_threshold):
                    grouped_overlaps[r1.name][group_num] = o1
                    grouped_overlaps[r2.name][group_num] = o2
                    group_num +=1
    return grouped_overlaps


def align_boundaries_with_overlap_sections(regions, remove_gaps=True):
    for rname, r in regions.region_dict.items():
        overlap_list = [o for o in list(regions.overlaps) if rname in o]
        if len(overlap_list) != 0:
            arbitrary_o = regions.overlaps[overlap_list[0]][rname]
            regions.region_dict[rname] = roll_boundary_to_align_with_overlap(
                r, arbitrary_o, remove_gaps=remove_gaps
            )
            
def roll_boundary_to_align_with_overlap(r, oidx, remove_gaps=True):
    roll_idx = -consecutive_lists(oidx[0], mod=r.lons_c.size)[0][0]
    new_oidx = {onum: np.mod(np.array(o) + roll_idx, r.lons_c.size) for (onum, o) in oidx.items()}
    
    r.lons_c = np.roll(r.lons_c, roll_idx)
    r.lats_c = np.roll(r.lats_c, roll_idx)
    
    if remove_gaps:
        for onum, o in new_oidx.items():
            gaps = np.array([i for i in range(np.min(o), np.max(o)+1) if i not in o])
            if any(gaps):
                r.lons_c[gaps] = np.nan
                r.lats_c[gaps] = np.nan

    nan_idx = np.isnan(r.lons_c) | np.isnan(r.lats_c)
    r.lons_c = r.lons_c[~nan_idx] 
    r.lats_c = r.lats_c[~nan_idx]
        
    return r


import geopandas as gpd
from shapely.geometry import Polygon
import regionmask

import sectionate as sec
import numpy as np

from .utilities import *

def get_region_boundary_grid_indices(lons, lats, grid):
    """Find boundary coordinates and grid indices that approximate a polygon.

    ARGUMENTS
    ---------
    lons : list or np.ndarray of longitudes
    lats : list or np.ndarray of latitudes
    grid : `xgcm.Grid` instance
    
    RETURNS
    -------
    (i_c, j_c, lons_c, lats_c, lons_uv, lats_uv)

    i_c : "X"-axis grid indices of corner points
    j_c : "Y"-axis grid indices of corner points
    lons_c : longitudes of corner points
    lats_c : latitudes of corner points
    lons_uv : longitudes of (u,v) velocity faces
    lats_uv : latitudes of (u,v) velocity faces
    """
    if (lons[0], lats[0]) != (lons[-1], lats[-1]):
        lons, lats = loop(lons), loop(lats)
        
    i_c, j_c, lons_c, lats_c = sec.grid_section(
        grid,
        lons,
        lats,
        topology="MOM-tripolar"
    )
    lons_uv, lats_uv = sec.uvcoords_from_qindices(grid, i_c, j_c)

    return (i_c, j_c, lons_c, lats_c, lons_uv, lats_uv)

def mask_from_grid_boundaries(
    lons_c,
    lats_c,
    grid,
    along_boundary=False,
    coordnames={'h': ('geolon', 'geolat')},
    ):
    """Find mask bounded by a sequence of cell corner coordinates

    ARGUMENTS
    ---------
    lons_c [list or np.ndarray] -- cell corner longitudes
    lats_c [list or np.ndarray] -- cell corner latitudes
    grid [xgcm.Grid] -- ocean model grid
    
    RETURNS
    -------
    region_grid_mask : np.ndarray of bool type
    """
    
    Δlon = np.sum(np.diff(lons_c)[np.abs(np.diff(lons_c)) < 180])
    crs = 'epsg:4326'
    
    if along_boundary:
        polygon_geom = Polygon(zip(lons_c, lats_c))

        polygon = gpd.GeoDataFrame(index=[0], crs=crs, geometry=[polygon_geom])
        region_grid_mask = ~np.isnan(
            regionmask.mask_geopandas(
                polygon,
                grid._ds[coordnames['h'][0]],
                lat=grid._ds[coordnames['h'][1]],
                wrap_lon=False
            )
        )
        
    elif np.abs(Δlon) < 180.:
        lons = loop(lons_c)
        lats = loop(lats_c)
        wrapped_lons = wrap_continuously(lons)
        minlon = np.min(wrapped_lons)
        polygon_geom = Polygon(zip(np.mod(wrapped_lons-minlon, 360.), lats))

        polygon = gpd.GeoDataFrame(index=[0], crs=crs, geometry=[polygon_geom])
        region_grid_mask = ~np.isnan(
            regionmask.mask_geopandas(
                polygon,
                np.mod(grid._ds[coordnames['h'][0]]-minlon, 360.),
                lat=grid._ds[coordnames['h'][1]],
                wrap_lon='360'
            )
        )
        
    else:
        # Follow sectionate convention for orientations of polygons
        # on stereographic plane (relative to South Pole)
        s = np.sign(Δlon).astype(int)
        if s==-1:
            lons_c = lons_c[::-1]
            lats_c = lats_c[::-1]
            s = 1

        min_idx = np.argmin(lons_c)

        lons = np.roll(lons_c, -min_idx)
        lats = np.roll(lats_c, -min_idx)

        lons = np.append(lon_mod(lons[-1], lons[0]), lons)
        lats = np.append(lats[-1], lats)

        diffs = s*(lons[np.newaxis, :] - lons[:, np.newaxis])
        diffs[np.tril_indices(lons.size)]*=-1
        single_valued = ~np.any(diffs < 0, axis=1)

        roll_idx = np.argmax(single_valued[::s])
        lons = np.roll(lons[::s], -roll_idx)[::s]
        lats = np.roll(lats[::s], -roll_idx)[::s]
        lons[::s][-roll_idx:] = lons[::s][-roll_idx:]

        # Make sure that we get everything to the South
        min_idx = np.argmin(lons)
        max_idx = np.argmax(lons)
        lons = np.append(
            lons, [
                lons[max_idx]+10,
                lons[max_idx]+10,
                lons[min_idx]-10,
                lons[min_idx]-10
            ]
        )
        lats = np.append(
            lats, [
                lats[max_idx],
                -90,
                -90,
                lats[min_idx]
            ]
        )
        
        polygon_geom = Polygon(zip(lons, lats))
        polygon = gpd.GeoDataFrame(index=[0], crs=crs, geometry=[polygon_geom])
        region_grid_mask = ~np.isnan(
            regionmask.mask_geopandas(
                polygon,
                grid._ds[coordnames['h'][0]],
                lat=grid._ds[coordnames['h'][1]],
                wrap_lon=False
            )
        )
    
    return region_grid_mask

def wrap_continuously(x, limit_discontinuity=180.):
    new_x = x.copy()
    for i in range(len(new_x)-1):
        if new_x[i+1]-new_x[i] >= 180.:
            new_x[i+1] -= 360.
        elif new_x[i+1]-new_x[i] < -180:
            new_x[i+1] += 360.
    return new_x



import contourpy
import xarray as xr
import numpy as np

from .utilities import loop, unique_list
from sectionate.transports import check_symmetric

def grid_boundaries_from_mask(grid, mask):
    """Use `contourpy` to get indices of cell corners that bound the mask

    Returns lists with a common length that is the number of discrete contours
    that bound the mask.
    
    ARGUMENTS
    ---------
    grid : `xgcm.Grid` instance
    mask : `xr.DataArray` instance of type `bool`
    
    RETURNS
    -------
    i_c_list, j_c_list, lons_c_list, lats_c_list
    """
    
    symmetric = check_symmetric(grid)
    
    contours = (
        contourpy.contour_generator(
            np.arange(-1, mask.shape[1]+1),
            np.arange(-1, mask.shape[0]+1),
            np.pad(mask.values, np.array([1,1]))
        )
        .create_contour(0.5)
    )
    i_c_list = []
    j_c_list = []
    lons_c_list = []
    lats_c_list = []
    for c in contours:
        i_c, j_c = c[:-1,0], c[:-1,1]
        
        i_c_new, j_c_new = i_c.copy(), j_c.copy()
        
        i_inc = np.roll(i_c, -1)-i_c
        j_inc = np.roll(j_c, -1)-j_c
        
        i_c_new[(i_c%1)==0.0] = (i_c - (i_inc<0))[(i_c%1)==0.0] + symmetric
        j_c_new[(j_c%1)==0.0] = (j_c - (j_inc<0))[(j_c%1)==0.0] + symmetric
        i_c_new[(i_c%1)==0.5] = np.floor(i_c[(i_c%1)==0.5]) + symmetric
        j_c_new[(j_c%1)==0.5] = np.floor(j_c[(j_c%1)==0.5]) + symmetric
        
        i_c_new, j_c_new = loop(i_c_new).astype(np.int64), loop(j_c_new).astype(np.int64)
                        
        i_c_list.append(i_c_new)
        j_c_list.append(j_c_new)
        
        idx = {
            grid.axes["X"].coords["outer"]:xr.DataArray(i_c_new, dims=("pt",)),
            grid.axes["Y"].coords["outer"]:xr.DataArray(j_c_new, dims=("pt",))
        }
        lons_c_list.append(grid._ds["geolon_c"].isel(idx).values[:-1])
        lats_c_list.append(grid._ds["geolat_c"].isel(idx).values[:-1])
        
    return i_c_list, j_c_list, lons_c_list, lats_c_list


import numpy as np
import xarray as xr

from .region import Region, GriddedRegion, BoundedRegion, open_gr
from .boundaries import grid_boundaries_from_mask
from .overlaps import *
from .utilities import *

import os
from pathlib import Path

class Regions():
    """
    A dictionary of polygonal regions defined by a list or array of geographical coordinates.
    """
    def __init__(
        self,
        region_dict,
        name=None
        ):
        """
        Create a `Regions` object from a dictionary mapping region names to `Region` instances.

        PARAMETERS
        ----------
        region_dict : dictionary mapping region `name` (str) to `Region` instance
        name : str or None (default: None)
            Overarching name of the collection of regions

        RETURNS
        -------
        `Regions` instance

        Examples
        --------
        >>> lons, lats = np.array([-80., -66., -65.]), np.array([ 26.,  18.,  32.])
        >>> region = reg.Region('Bermuda Triangle', lons, lats)
        >>> regions = reg.Regions({region.name: region})
        """
        if type(region_dict) == dict:
            self.region_dict = region_dict
        else:
            raise NameError("Must provide `regions_dict` to initialize.")
        if name is not None:
            self.name = name
    
    def find_all_overlaps(self, closeness_threshold=5.e3, face_indices=False):
        """Finds indices of `self.overlaps` between region boundaries

        Combined with `sectionate`, this can be used to determine transports across
        shared boundaries between two adjacent regions.
        """
        self.overlaps = {}
        for i, (r1name, r1) in enumerate(self.region_dict.items()):
            for j, (r2name, r2) in enumerate(self.region_dict.items()):
                if r1name<r2name:
                    overlaps = find_indices_of_overlaps(
                        r1,
                        r2,
                        closeness_threshold=closeness_threshold,
                        face_indices=face_indices
                    )
                    if len(overlaps[r1name]):
                        self.overlaps[sorted_tuple((r1name, r2name))] = overlaps
                
    def copy(self, remove_duplicate_points=False):
        """
        Returns a copy of the Regions.

        PARAMETERS
        ----------
        remove_duplicate_points : bool
            Default: False. If True, prunes any duplicate points from the input arrays (lons, lats).
            
        RETURNS
        ----------
        region_copy : `regionate.regions.Regions` type
            Copy of the region.
        """
        return Regions({
            r.name: r.copy(remove_duplicate_points=remove_duplicate_points)
            for r in self.region_dict.values()
        })
    
class GriddedRegions(Regions):
    """
    A dictionary of named polygonal regions that exactly conform to the velocity faces of a C-grid ocean model.
    """
    def __init__(
        self,
        region_dict,
        grid,
        name=None,
        ):
        """
        Create a `GriddedRegions` object from a dictionary mapping region names to `GriddedRegion` instances.

        PARAMETERS
        ----------
        region_dict : dictionary mapping region `name` (str) to `Region`, `GriddedRegion`, or `BoundedRegion` instance
        name : str or None (default: None)
            Overarching name of the collection of regions

        RETURNS
        -------
        `GriddedRegions` instance
        """
        self.grid = grid
        
        super().__init__(region_dict, name=name)
        try:
            if all([type(v) in [Region, GriddedRegion, BoundedRegion] for v in region_dict.values()]):
                super().__init__(region_dict, name=name)
            else:
                raise NameError("""Values in `region_dict` dictionary must be instances of
                `Region`, `GriddedRegion`, or `BoundedRegion`.""")
        except:
            raise NameError("Must provide valid `region_dict` dictionary to initialize.")

    def to_grs(self, path):
        """
        TO DO
        """
        
        # Create .grs file directory
        grs_path = f"{path}{self.name.replace(' ','_')}.grs"
        Path(grs_path).mkdir(parents=True, exist_ok=True)

        # Write grid dataset (without variables) to NetCDF file
        grid = self.grid
        grid_path = f"{grs_path}/grid.nc"
        grid._ds.drop_vars([v for v in grid._ds.data_vars]).to_netcdf(grid_path)

        # Write boundary information for each region to separate .gr file directory
        for region_name, region in self.region_dict.items():
            region.to_gr(f"{grs_path}/")

class MaskRegions(GriddedRegions):
    """
    A dictionary of polygonal regions that exactly conform to the velocity faces bounding a mask in a C-grid ocean model.
    """
    def __init__(
        self,
        mask,
        grid,
        name=None,
        ):
        """
        Create a `MaskRegions` object from a mask and accompanying `xgcm.Grid` instance.

        PARAMETERS
        ----------
        mask : None or xr.DataArray (default: None)
            If None, does not apply any mask.
        grid : `xgcm.Grid` instance
        name : str or None (default: None)
            Overarching name of the collection of regions

        RETURNS
        -------
        `GriddedRegions` instance
        """
        
        if any([c not in grid._ds.coords for c in ["geolon_c", "geolat_c"]]):
            raise ValueError("grid._ds must contain coordinates of grid cell corners, named 'geolon_c' and 'geolat_c'.")

        self.grid = grid
        self.mask = mask
        
        i_c_list, j_c_list, lons_c_list, lats_c_list = grid_boundaries_from_mask(
            self.grid,
            mask
        )

        region_dict = {
            r_num: GriddedRegion(
                str(r_num),
                lons_c,
                lats_c,
                self.grid,
                mask=mask,
                ij=(i_c,j_c)
            )
            for r_num, (i_c, j_c, lons_c, lats_c)
            in enumerate(zip(i_c_list, j_c_list, lons_c_list, lats_c_list))
        }
        super().__init__(region_dict, grid, name=name)

def open_grs(path, ds_to_grid):
    """
    Creates a GriddedRegions instance from a .grs file format

    PARAMETERS
    ----------
    path : str
        path to .grs directory
    ds_to_grid : function
        a function whose only argument is an xr.Dataset object.
        This function turns `xr.open_dataset(f"{path}/grid.nc")`
        into a corresponding `xgcm.Grid` object.

    RETURNS
    -------
    `GriddedRegions` instance
    """
    name = path.split('/')[-1][:-4]
    grid = ds_to_grid(xr.open_dataset(f"{path}/grid.nc"))
    grs_files = [f for f in os.listdir(f"{path}/") if f!='grid.nc']
    region_dict = {}
    for gr_file in grs_files:
        gr_name = gr_file[:-3]
        region_dict[gr_name] = open_gr(f"{path}/{gr_file}", ds_to_grid)

    return GriddedRegions(region_dict, grid, name=name)

def sorted_tuple(s):
    """Sort tuples (by integer)"""
    return tuple(sorted(s, key=int))


import xarray as xr

def check_global_coverage(regions):
    """Check whether the masks in the `Regions` instance are non-overlapping and provide complete global coverage"""
    total_mask = xr.zeros_like(list(regions.region_dict.values())[0].mask)
    for r in regions.region_dict.values():
        total_mask += r.mask
    if (total_mask == 1).sum() != total_mask.size:
        ValueError(f"Region {r.name} has incomplete or imperfect global coverage.")